from canvasapi.course import Course
from canvasapi.file import File
from canvasapi.folder import Folder

from .util import get_canvas_object
from ..resources import FileData
from ..our_logging import get_logger

logger = get_logger()

from ..our_logging import get_logger

logger = get_logger()
DEFAULT_CANVAS_FOLDER = 'deployed_files'


def get_file(course: Course, name: str) -> File:
    return get_canvas_object(course.get_files, 'display_name', name)


def get_canvas_folder(course: Course, folder_name: str, parent_folder_path="") -> Folder:
    """
    Retrieves an object representing a digital folder in Canvas.
    If the folder does not exist, it is created.
    """
    folder = get_canvas_object(course.get_folders, 'name', folder_name)
    if folder is not None:
        return folder

    logger.debug(f"Creating folder: {folder_name}")
    return course.create_folder(name=folder_name, parent_folder_path=parent_folder_path, hidden=True)


def deploy_file(course: Course, data: FileData):
    canvas_folder = data.get('canvas_folder') or DEFAULT_CANVAS_FOLDER
    folder = get_canvas_folder(course, canvas_folder)
    file_id = folder.upload(data['path'])[1]['id']
    return course.get_file(file_id), None


def lookup_file(course: Course, name: str):
    # TODO - do we need to search against the folder also?
    # If so, include the folder in the name (e.g. /unit1/stuff/file.txt)
    # and parse it out here
    return get_file(course, name)



from canvasapi.course import Course
from canvasapi.quiz import Quiz

from .util import get_canvas_object, update_group_name_to_id, ResourceNotFoundException

from ..our_logging import get_logger

logger = get_logger()

def get_quiz(course: Course, title: str) -> Quiz:
    return get_canvas_object(course.get_quizzes, "title", title)


def debug_quiz_creation(canvas_quiz: Quiz, course: Course, data):
    new_settings = {"title": data["title"]}

    for key, value in zip(data.keys(), data.values()):
        new_settings[key] = value
        logger.debug(f"Attempting with {key}: {value}")
        try:
            canvas_quiz = course.create_quiz(quiz=new_settings)
        except Exception as ex:
            logger.exception(f"Failed on key: {key}, value: {value}")
            raise ex
        canvas_quiz.delete()
    return canvas_quiz


def create_quiz(course: Course, data: dict, name: str):
    logger.debug(f"Creating canvas quiz {name}...")
    try:
        canvas_quiz = course.create_quiz(quiz=data)
    except Exception as ex:
        logger.exception(f"Error creating canvas quiz {name}")

        # Perhaps the quiz was partially created, and then the program crashed
        if canvas_quiz := get_quiz(course, name):
            logger.warning(f"Attempting to edit partially created quiz {name}...")
            try:
                canvas_quiz.edit(quiz=data)
            except Exception as ex:
                logger.exception("Failed to edit quiz")
                raise ex
        else:
            logger.error("Quiz was not created")
            logger.error("Attempting to debug quiz creation")
            canvas_quiz = debug_quiz_creation(canvas_quiz, course, data)
    return canvas_quiz


def check_quiz(canvas_quiz, name: str):
    """
    Checks if quiz has submissions and throws a warning with link to quiz.
    """
    if any(canvas_quiz.get_submissions()):
        return f"Quiz {name} has submissions. See {canvas_quiz.html_url} to save quiz."
    return None


def replace_questions(quiz: Quiz, questions: list[dict]):
    """
    Deletes all questions in a quiz, and replaces them with new questions.
    """
    logger.debug(f"Replacing questions ... ")
    for quiz_question in quiz.get_questions():
        quiz_question.delete()
    for question in questions:
        quiz.create_question(question=question)


def deploy_quiz(course: Course, quiz_data: dict) -> Quiz:
    name = quiz_data['title']

    update_group_name_to_id(course, quiz_data)

    if canvas_quiz := get_quiz(course, name):
        canvas_quiz: Quiz
        warning = check_quiz(canvas_quiz, name)
        canvas_quiz.edit(quiz=quiz_data)
    else:
        canvas_quiz = create_quiz(course, quiz_data, name)
        warning = None

    replace_questions(canvas_quiz, quiz_data['questions'])
    canvas_quiz.edit()

    return canvas_quiz, warning


def lookup_quiz(course: Course, quiz_name: str) -> Quiz:
    canvas_quiz = get_quiz(course, quiz_name)
    if not canvas_quiz:
        raise ResourceNotFoundException(f'Quiz {quiz_name} not found')
    return canvas_quiz



import cssutils
from bs4 import BeautifulSoup
from ..util import parse_soup_from_xml


def get_style(soup):
    style = ''

    for tag in soup.find_all("style"):
        content = tag.text.strip()
        if content not in style:
            style += content + ' '
        tag.decompose()

    return style, soup


def parse_css(css):
    css_parser = cssutils.CSSParser()
    stylesheet = css_parser.parseString(css)
    styles = {}
    for rule in stylesheet:
        if rule.type == rule.STYLE_RULE:
            selector = rule.selectorText
            properties = {prop.name: prop.value for prop in rule.style}
            styles[selector] = properties
    return styles


def apply_inline_styles(html, styles):
    soup = parse_soup_from_xml(html)
    for selector, properties in styles.items():
        for tag in soup.select(selector):
            style_string = "; ".join([f"{prop}: {value}" for prop, value in properties.items()])
            existing_style = tag.get('style', '')
            tag['style'] = existing_style + ((existing_style and '; ') or '') + style_string
    return str(soup)


def bake_css(soup: BeautifulSoup, global_css: str):
    css, soup = get_style(soup)
    css = parse_css(global_css + css)
    soup = apply_inline_styles(str(soup), css)
    return soup


if __name__ == '__main__':
    # Example usage
    html_content = '''<p>Go team</p>'''
    css_content = '''p { font-size: 200px; }'''

    # Parse the CSS and HTML
    parsed_styles = parse_css(css_content)
    styled_html = apply_inline_styles(html_content, parsed_styles)
    print(styled_html)



import re
import textwrap
from xml.etree.ElementTree import Element

from bs4 import NavigableString
import markdown as md
from markdown import Extension
from markdown.extensions.codehilite import makeExtension as makeCodehiliteExtension
from markdown.inlinepatterns import BACKTICK_RE, BacktickInlineProcessor

from .inline_math import InlineMathExtension
from ..util import parse_soup_from_xml


class BlackInlineCodeProcessor(BacktickInlineProcessor):
    def handleMatch(self, m: re.Match[str], data: str) -> tuple[Element | str, int, int]:
        el, start, end = super().handleMatch(m, data)
        el.attrib['style'] = 'color: #000000'
        return el, start, end


class BlackInlineCodeExtension(Extension):
    def extendMarkdown(self, md):
        # We use 'backtick' and 190 which are the same values
        # used in markdown.inlinepatterns.py to register the original
        # BacktickInlineCodeProcessor.
        # By reusing the same name, it overrides the original processor with ours
        md.inlinePatterns.register(BlackInlineCodeProcessor(BACKTICK_RE), 'backtick', 190)


def process_markdown_text(text: str) -> str:
    dedented = textwrap.dedent(text)

    html = md.markdown(dedented, extensions=[
        'fenced_code',
        'tables',
        'attr_list',

        # This embeds the highlight style directly into the HTML
        # instead of using CSS classes
        makeCodehiliteExtension(noclasses=True),

        # This preserves \(...\) inline math expressions
        #  so Canvas will render them with MathJax
        InlineMathExtension(),

        # This forces the color of inline code to be black
        # as a workaround for Canvas's super-ugly default red :P
        BlackInlineCodeExtension(),
        # TODO - can we solve this with baked-in CSS?

        # TODO - add support for tilde => <del> (strikethrough) (look for extension)
        #  or maybe look for a github-flavored-markdown extension
    ])
    return html


def _process_markdown(parent, excluded: list[str]):
    children = list(parent.children)
    for tag in children:
        if tag.name in excluded:
            continue

        if isinstance(tag, NavigableString):
            tag.replace_with(parse_soup_from_xml(process_markdown_text(tag.text)))
        else:
            _process_markdown(tag, excluded)


def process_markdown(text: str, excluded: list[str]) -> str:
    """
    Process Markdown text and return XML text

    This purpose of this function is only the Markdown to XML step
    Custom XML/HTML tags should be handled by the XML processor
    This function simply converts all Markdown formatting to HTML

    This function processes Markdown in ALL XML/HTML tags
    (including nested tags) except those listed in `excluded`.

    :param text: the Markdown text to process
    :param excluded: a list of tag names to exclude; their contents are left untouched
    :returns: The XML/HTML text
    """
    soup = parse_soup_from_xml(text)
    _process_markdown(soup, excluded)
    return str(soup)



from canvasapi.course import Course
from canvasapi.module import Module, ModuleItem

from .assignment import get_assignment
from .file import get_file
from .page import get_page
from .quiz import get_quiz
from .util import get_canvas_object


def _get_module(course: Course, name: str) -> Module:
    return get_canvas_object(course.get_modules, 'name', name)


def _get_module_item(module: Module, item: dict) -> ModuleItem | None:
    for module_item in module.get_module_items():
        if item['title'] == module_item.title:
            return module_item
    return None


def _delete_obsolete_module_items(module: Module, module_items: list[dict]):
    keepers = set(item['title'] for item in module_items)

    for module_item in module.get_module_items():
        if module_item.title not in keepers:
            module_item.delete()


def _add_canvas_id(course: Course, item: dict):
    # Add the content_id or page_url as described in
    # https://canvas.instructure.com/doc/api/modules.html#method.context_module_items_api.create

    if item['type'] in ['ExternalUrl', 'SubHeader']:
        # content_id not required
        return

    if item['type'] == 'Page':
        item['page_url'] = get_page(course, item['title']).url

    elif item['type'] == 'Quiz':
        item['content_id'] = get_quiz(course, item['title']).id

    elif item['type'] == 'Assignment':
        item['content_id'] = get_assignment(course, item['title']).id

    elif item['type'] == 'File':
        item['content_id'] = get_file(course, item['title']).id

    else:
        raise NotImplementedError('Module item of type ' + item['type'])


def _create_or_update_module_items(course: Course, module: Module, module_items: list[dict]):
    _delete_obsolete_module_items(module, module_items)

    # TODO - make sure the order of items matches the order in the XML

    for index, item in enumerate(module_items):

        _add_canvas_id(course, item)

        if module_item := _get_module_item(module, item):
            module_item.edit(module_item=item)
        else:
            module.create_module_item(module_item=item)


def deploy_module(course: Course, module_data: dict) -> tuple[Module, str|None]:
    name = module_data["name"]

    if canvas_module := _get_module(course, name):
        if 'published' not in module_data:
            module_data['published'] = canvas_module.published
        canvas_module.edit(module=module_data)
    else:
        canvas_module = course.create_module(module=module_data)

    _create_or_update_module_items(course, canvas_module, module_data.get('items', []))

    return canvas_module, None


lookup_module = _get_module



# TODO
# <assignment_groups>
# | Group | Weight |
# |-------|--------|
# | Homework | 10%  |
# | Projects | 40%  |
# | Exams    | 10%  |
# </assignment_groups>
# Parse that table into a dictionary
# str: int

# To deploy:
# Using the course.update() method,
# set the 'apply_assignment_group_weights' field to True
#
# # Get all assignment groups for the course
# assignment_groups = course.get_assignment_groups()
#
# Create the missing groups
#
# Map all names to IDs
#
# # Define the weights you want to set
# # Note: This should be a dict mapping group ids to their corresponding weights
# weights = {
#     'Homework': 30,
#     'Projects': 50,
#     'Exams': 20,
# }
#
# # Update the weights for each assignment group
# for group in assignment_groups:
#     group_name = group.name  # is this field available?
#     if group_name in weights:
#         new_weight = weights[group_name]
#         updated_group = group.edit(assignment_group={'group_weight': new_weight})



from markdown.extensions import Extension

from markdown.postprocessors import Postprocessor
from markdown.preprocessors import Preprocessor, HtmlBlockPreprocessor
from bs4 import BeautifulSoup
from bs4.element import Tag

from typing import Protocol

from mdxcanvas.xml_processing.inline_styling import get_style, parse_css, apply_inline_styles


# Make a Protocol for any tag processor, it should take a Tag and return a Tag
# This way we can define a type hint for the tag_processors dictionary
class TagProcessor(Protocol):
    def __call__(self, tag: Tag) -> Tag:
        ...


class CustomHTMLBlockTagProcessor(HtmlBlockPreprocessor):
    tag_processors: dict[str, TagProcessor]

    def __init__(self, md, tag_processors: dict[str, TagProcessor]):
        super().__init__(md)
        self.custom_tag_processors = tag_processors

    def run(self, lines: list[str]) -> list[str]:
        joined = "\n".join(lines)
        if "<" not in joined or ">" not in joined:
            return lines
        soup = BeautifulSoup(joined, "html.parser")
        for tag in soup.find_all():
            if tag.name in self.custom_tag_processors:
                processor = self.custom_tag_processors[tag.name]
                new_tag = processor(tag)
                tag.replace_with(new_tag)
        return str(soup).split("\n")


class PrintLinesPreprocessor(Preprocessor):
    """
    Used for debugging the state of the document
    in the middle of the preprocessor chain
    """
    def __init__(self, md, label):
        super().__init__(md)
        self.label = label

    def run(self, lines: list[str]) -> list[str]:
        print(f'START LINES {self.label} ----------------------------------------')
        for line in lines:
            print(line, end='' if line.endswith('\n') else '\n')
        print(f'END LINES   {self.label} ----------------------------------------')
        return lines


class CustomTagExtension(Extension):
    def __init__(self, tag_processors: dict[str, TagProcessor]):
        super().__init__()
        self.tag_processors = tag_processors

    def extendMarkdown(self, md):
        # When registering the CustomHTMLBlockTagProcessor, we use a priority of 22
        # which is two more than the original priority for 'html_block'
        # (i.e. will run BEFORE the 'html_block' processor)
        md.preprocessors.register(
            CustomHTMLBlockTagProcessor(md, self.tag_processors),
            'custom_tag', 22
        )


class BakedCSSPostProcessor(Postprocessor):
    def __init__(self, global_css):
        super().__init__()
        self.global_css = global_css

    def run(self, text):
        soup = BeautifulSoup(text, 'html.parser')
        css, soup = get_style(soup)
        css = parse_css(self.global_css + css)
        soup = apply_inline_styles(str(soup), css)
        return str(soup)


class BakedCSSExtension(Extension):
    def __init__(self, global_css: str = ''):
        super().__init__()
        self.global_css = global_css

    def extendMarkdown(self, md):
        # By default something is at 20 and 30 so we chose 7 so it runs last
        md.postprocessors.register(
            BakedCSSPostProcessor(self.global_css),
            'baked-css', 7
        )


import json

from datetime import datetime
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Callable

import pytz
from canvasapi.canvas_object import CanvasObject
from canvasapi.course import Course

from .algorithms import linearize_dependencies
from .checksums import MD5Sums, compute_md5
from .file import deploy_file, lookup_file
from .syllabus import deploy_syllabus, lookup_syllabus
from .util import get_canvas_uri
from .zip import deploy_zip, lookup_zip, predeploy_zip
from .quiz import deploy_quiz, lookup_quiz, check_quiz
from .page import deploy_page, lookup_page
from .assignment import deploy_assignment, lookup_assignment
from .module import deploy_module, lookup_module

from ..resources import CanvasResource, iter_keys
from ..our_logging import log_warnings, get_logger

logger = get_logger()


def deploy_resource(course: Course, resource_type: str, resource_data: dict) -> tuple[CanvasObject, str|None]:
    deployers: dict[str, Callable[[Course, dict], tuple[CanvasObject, str|None]]] = {
        'zip': deploy_zip,
        'file': deploy_file,
        'page': deploy_page,
        'quiz': deploy_quiz,
        'assignment': deploy_assignment,
        'module': deploy_module,
        'syllabus': deploy_syllabus
    }

    if (deploy := deployers.get(resource_type, None)) is None:
        raise Exception(f'Deployment unsupported for resource of type {resource_type}')

    try:
        deployed, warning = deploy(course, resource_data)
    except:
        logger.error(f'Failed to deploy resource: {resource_type} {resource_data}')
        raise

    if deployed is None:
        raise Exception(f'Resource not found: {resource_type} {resource_data}')

    return deployed, warning


def lookup_resource(course: Course, resource_type: str, resource_name: str) -> CanvasObject:
    finders: dict[str, Callable[[Course, str], CanvasObject]] = {
        'zip': lookup_zip,
        'file': lookup_file,
        'page': lookup_page,
        'quiz': lookup_quiz,
        'assignment': lookup_assignment,
        'module': lookup_module,
        'syllabus': lookup_syllabus
    }

    if (finder := finders.get(resource_type, None)) is None:
        raise Exception(f'Lookup unsupported for resource of type {resource_type}')

    found = finder(course, resource_name)

    if found is None:
        raise Exception(f'Resource not found: {resource_type} {resource_name}')

    return found


def update_links(course: Course, data: dict, resource_objs: dict[tuple[str, str], CanvasObject]) -> dict:
    text = json.dumps(data)
    logger.debug(f'Updating links in {text}')

    for key, rtype, rname, field in iter_keys(text):
        logger.debug(f'Processing key: {key}, {rtype}, {rname}, {field}')

        if (rtype, rname) not in resource_objs:
            logger.info(f'Retrieving {rtype} {rname}')
            resource_objs[rtype, rname] = lookup_resource(course, rtype, rname)

        obj = resource_objs[rtype, rname]
        if field == 'uri':
            repl_text = get_canvas_uri(obj)
        else:
            repl_text = str(getattr(obj, field))
        text = text.replace(key, repl_text)

    return json.loads(text)


def make_iso(date: datetime | str | None, time_zone: str) -> str:
    if isinstance(date, datetime):
        return datetime.isoformat(date)
    elif isinstance(date, str):
        try_formats = [
            "%b %d, %Y, %I:%M %p",
            "%b %d %Y %I:%M %p",
            "%Y-%m-%dT%H:%M:%S",
            "%Y-%m-%dT%H:%M:%S%z"
        ]
        for format_str in try_formats:
            try:
                parsed_date = datetime.strptime(date, format_str)
                if parsed_date.tzinfo:
                    return datetime.isoformat(parsed_date)
                break
            except ValueError:
                pass
        else:
            raise ValueError(f"Invalid date format: {date}")

        # Convert the parsed datetime object to the desired timezone
        to_zone = pytz.timezone(time_zone)
        localized_date = to_zone.localize(parsed_date)
        return datetime.isoformat(localized_date)
    else:
        raise TypeError("Date must be a datetime object or a string")


def fix_dates(data, time_zone):
    for attr in ['due_at', 'unlock_at', 'lock_at', 'show_correct_answers_at']:
        if attr not in data:
            continue

        datetime_version = datetime.fromisoformat(make_iso(data[attr], time_zone))
        utc_version = datetime_version.astimezone(pytz.utc)
        data[attr] = utc_version.isoformat()


def get_dependencies(resources: dict[tuple[str, str], CanvasResource]) -> dict[tuple[str, str], list[str]]:
    """Returns the dependency graph in resources. Adds missing resources to the input dictionary."""
    deps = {}
    missing_resources = []
    for key, resource in resources.items():
        deps[key] = []
        text = json.dumps(resource)
        for _, rtype, rname, _ in iter_keys(text):
            resource_key = (rtype, rname)
            deps[key].append(resource_key)
            if resource_key not in resources:
                missing_resources.append(resource_key)

    for rtype, rname in missing_resources:
        resources[rtype, rname] = CanvasResource(type=rtype, name=rname, data=None)

    return deps


def predeploy_resource(rtype: str, resource_data: dict, timezone: str, tmpdir: Path) -> dict:
    fix_dates(resource_data, timezone)

    predeployers: dict[str, Callable[[dict, Path], dict]] = {
        'zip': predeploy_zip
    }

    if (predeploy := predeployers.get(rtype)) is not None:
        logger.debug(f'Predeploying {rtype} {resource_data}')
        resource_data = predeploy(resource_data, tmpdir)

    return resource_data


def deploy_to_canvas(course: Course, timezone: str, resources: dict[tuple[str, str], CanvasResource]):
    resource_dependencies = get_dependencies(resources)
    logger.debug(f'Dependency graph: {resource_dependencies}')

    resource_order = linearize_dependencies(resource_dependencies)
    logger.debug(f'Order of deployment: {resource_order}')

    warnings = []
    logger.info('Beginning deployment to Canvas')
    with MD5Sums(course) as md5s, TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)

        resource_objs: dict[tuple[str, str], CanvasObject] = {}
        for resource_key in resource_order:
            try:
                logger.debug(f'Processing {resource_key}')
                resource = resources[resource_key]

                rtype = resource['type']
                rname = resource['name']
                logger.info(f'Processing {rtype} {rname}')
                if (resource_data := resource.get('data')) is not None:
                    # Deploy resource using data
                    resource_data = predeploy_resource(rtype, resource_data, timezone, tmpdir)
                    resource_data = update_links(course, resource_data, resource_objs)

                    stored_md5 = md5s.get(resource_key)
                    current_md5 = compute_md5(resource_data)

                    if current_md5 != stored_md5:
                        # Create the resource
                        logger.info(f'Deploying {rtype} {rname}')
                        resource_obj, warning = deploy_resource(course, rtype, resource_data)
                        if warning:
                            warnings.append(warning)
                        resource_objs[resource_key] = resource_obj
                        md5s[resource_key] = current_md5
            except:
                logger.error(f'Error deploying resource {rtype} {rname}')
                raise

        if warnings:
            log_warnings(warnings)
    # Done!



from pathlib import Path
from typing import Callable

from .assignment_tags import AssignmentTagProcessor
from .syllabus_tags import SyllabusTagProcessor
from ..resources import ResourceManager
from ..util import parse_soup_from_xml
from ..xml_processing.tag_preprocessors import make_image_preprocessor, make_file_preprocessor, \
    make_zip_preprocessor, make_include_preprocessor, make_link_preprocessor, make_markdown_page_preprocessor
from ..xml_processing.quiz_tags import QuizTagProcessor
from ..xml_processing.page_tags import PageTagProcessor
from ..xml_processing.module_tags import ModuleTagProcessor


def _walk_xml(tag, tag_processors):
    if not hasattr(tag, 'children'):
        return
    for child in tag.children:
        if hasattr(child, 'name') and child.name in tag_processors:
            processor = tag_processors[child.name]
            processor(child)
        _walk_xml(child, tag_processors)


def preprocess_xml(
        parent: Path,
        text: str,
        resources: ResourceManager,
        process_file: Callable
) -> str:
    """
    Preprocess the XML/HTML text to handle special content tags
    e.g. links, images, files, includes, etc.

    Returns modified XML that uses local IDs in the links.
    These IDs will be replaced with real Canvas IDs during deployment.
    """
    tag_preprocessors = {
        'img': make_image_preprocessor(parent, resources),
        'file': make_file_preprocessor(parent, resources),
        'zip': make_zip_preprocessor(parent, resources),
        'include': make_include_preprocessor(parent, process_file),
        'course-link': make_link_preprocessor(),
        'md-page': make_markdown_page_preprocessor(parent, process_file),
    }

    soup = parse_soup_from_xml(text)
    _walk_xml(soup, tag_preprocessors)

    return str(soup)


def process_canvas_xml(resources: ResourceManager, text: str):
    """
    Process XML/HTML text into a DTOs that represent
    the content to be deployed to Canvas.

    :param text: The XML/HTML text to be processed
    :returns: Populated ResourceManager
    """

    # -- Strategy --
    # The algorithm walks the tree and calls the appropriate processor on each tag
    # Each custom tag is processed by a bespoke processor
    # The tag processor returns Canvas JSON
    # If the tag is not processed (no assigned processor),
    #  the algorithm recurses on its children

    tag_processors = {
        'quiz': QuizTagProcessor(resources),
        'page': PageTagProcessor(resources),
        'assignment': AssignmentTagProcessor(resources),
        'module': ModuleTagProcessor(resources),
        'syllabus': SyllabusTagProcessor(resources)
    }

    soup = parse_soup_from_xml(text)
    _walk_xml(soup, tag_processors)

    return resources



import csv
import json
import re
from pathlib import Path
from typing import Any

import jinja2 as jj
from bs4.element import Tag

from .markdown_processing import process_markdown_text
from ..our_logging import get_logger
from ..util import parse_soup_from_xml, retrieve_contents

logger = get_logger()


def _tokenize(text: str, break_tags):
    soup = parse_soup_from_xml(text)
    current_section = ''

    for tag in soup.find_all():
        if tag.name in break_tags and current_section:
            yield current_section
            current_section = ''
        current_section += str(tag)

    if current_section:
        yield current_section


def _extract_headers(table: Tag) -> list[str]:
    return [th.text.strip() for th in table.find_all('th')]


def _extract_row_data(headers: list[str], row: Tag) -> dict:
    cells = row.find_all(['td', 'th'])
    if len(cells) != len(headers):
        logger.debug(f'Row data does not match headers: {row} ')
        return {}
    return {headers[i]: retrieve_contents(cells[i]) for i in range(len(headers))}


def _process_table(tag: Tag) -> list[dict]:
    headers = _extract_headers(tag)
    rows = (_extract_row_data(headers, tr) for tr in tag.find_all('tr')[1:])
    return [row for row in rows if row]


def _process_h1_table(tag: Tag) -> dict:
    headers = _extract_headers(tag)

    if headers == ['Key', 'Value']:
        rows = (_extract_row_data(headers, tr) for tr in tag.find_all('tr')[1:])
        return {
            row['Key']: row['Value']
            for row in rows
            if row
        }

    else:
        return _extract_row_data(headers, tag.find_all('tr')[1])


def _parse_h1_header_data(text: str):
    section_data = {}

    tokens = _tokenize(text, ['h1', 'h3', 'table'])
    for token in tokens:
        soup = parse_soup_from_xml(token)
        first_tag = soup.find()
        if first_tag.name == 'h1':
            section_data['Title'] = first_tag.text.strip()

        elif first_tag.name == 'h3':
            section_data[first_tag.text.strip()] = ''.join(str(tag) for tag in first_tag.next_siblings)

        elif first_tag.name == 'table':
            section_data |= _process_h1_table(first_tag)

    return section_data


def _parse_h2_section(text: str) -> tuple[str, list[dict]]:
    soup = parse_soup_from_xml(text)

    title_tag = soup.find('h2')
    section_name = title_tag.text.strip()

    table_tag = soup.find('table')
    section_data = _process_table(table_tag) if table_tag else []

    return section_name, section_data


def _parse_h1_section(h1_section: str) -> dict[str, Any]:
    tokens = list(_tokenize(h1_section, ['h1', 'h2']))

    h1_section_data: dict[str, Any] = _parse_h1_header_data(tokens[0])

    for token in tokens[1:]:
        section_name, h2_section_data = _parse_h2_section(token)
        h1_section_data[section_name] = h2_section_data

    return h1_section_data


def _read_multiple_tables(text: str):
    rows = []
    for h1_section in _tokenize(text, ['h1']):
        rows.append(_parse_h1_section(h1_section))
    return rows


def _read_single_table(html: str) -> list[dict]:
    soup = parse_soup_from_xml(html)
    table = soup.find('table')
    if table is None:
        return []
    return _process_table(table)


def _read_md_table(md_text: str) -> list[dict]:
    html = process_markdown_text(md_text)

    # Check if file contains header 1 or 2 tags, indicating multiple tables
    if re.search(r'<h[1|2]>', html):
        return _read_multiple_tables(html)
    else:
        return _read_single_table(html)


def _get_global_args(global_args_path: Path) -> dict:
    if '.json' in global_args_path.name:
        return json.loads(global_args_path.read_text())
    elif '.csv' in global_args_path.name:
        return dict(csv.DictReader(global_args_path.read_text().splitlines()))
    else:
        raise NotImplementedError('Global args file of type: ' + global_args_path.suffix)


def _get_args(args_path: Path) -> list[dict]:
    if args_path.suffix == '.json':
        return json.loads(args_path.read_text())

    elif args_path.suffix == '.csv':
        return list(csv.DictReader(args_path.read_text().splitlines()))

    elif args_path.suffix == '.md':
        return _read_md_table(args_path.read_text())

    else:
        raise NotImplementedError('Args file of type: ' + args_path.suffix)


def _render_template(template, **kwargs):
    jj_template = jj.Environment().from_string(template)
    kwargs |= dict(zip=zip, split_list=lambda x: x.split(';'))
    return jj_template.render(**kwargs)


def _process_template(template: str, arg_sets: list[dict]):
    return '\n'.join([_render_template(template, **args) for args in arg_sets])


def process_jinja(
        template: str,
        args_path: Path = None,
        global_args_path: Path = None,
        **kwargs
) -> str:
    arg_sets = _get_args(args_path) if args_path is not None else None

    if global_args_path:
        kwargs |= _get_global_args(global_args_path)

    if arg_sets is not None:
        arg_sets = [{**args, **kwargs} for args in arg_sets]
    else:
        arg_sets = [kwargs]

    return _process_template(template, arg_sets)



import argparse
import json
import logging
import os
from pathlib import Path
from typing import TypedDict

from canvasapi import Canvas
from canvasapi.course import Course

from .xml_processing.inline_styling import bake_css
from .util import parse_soup_from_xml
from .deploy.canvas_deploy import deploy_to_canvas
from .resources import ResourceManager
from .xml_processing.xml_processing import process_canvas_xml, preprocess_xml
from .text_processing.markdown_processing import process_markdown
from .text_processing.jinja_processing import process_jinja

from .our_logging import get_logger

logger = get_logger()

class CourseInfo(TypedDict):
    CANVAS_API_URL: str
    CANVAS_COURSE_ID: int
    LOCAL_TIME_ZONE: str


def read_content(input_file: Path) -> tuple[list[str], str]:
    return input_file.suffixes, input_file.read_text()


def is_jinja(content_type):
    return content_type[-1] == '.jinja'


def _post_process_content(xml_content: str, global_css: str) -> str:
    # - bake in CSS styles
    soup = parse_soup_from_xml(xml_content)
    xml_postprocessors = [
        lambda s: bake_css(s, global_css)
    ]
    for xml_post in xml_postprocessors:
        soup = xml_post(soup)

    return str(soup)


def process_file(
        resources: ResourceManager,
        parent_folder: Path,
        content: str,
        content_type: list[str],
        global_args_file: Path = None,
        args_file: Path = None,
        line_id: str = None,
        css_file: Path = None
) -> str:
    """
    Read a file and fully process the text content
    Process Markdown.
    Process content-modifying XML tags (e.g. img, or file, or zip, or include)
    Post-process the content (whole XML in, whole XML out, e.g. bake CSS)
    """
    if is_jinja(content_type):
        content = process_jinja(
            content,
            args_path=args_file,
            global_args_path=global_args_file,
            line_id=line_id
        )

    if '.md' in content_type:
        # Process Markdown
        excluded = ['pre', 'style']

        logger.debug('Processing Markdown')
        xml_content = process_markdown(content, excluded=excluded)

    else:
        logger.debug('No Markdown processing needed')
        xml_content = content

    # Preprocess XML
    logger.debug('Processing XML')

    def load_and_process_file_contents(parent: Path, content: str, content_type: list[str], **kwargs) -> str:
        return process_file(resources, parent, content, content_type, global_args_file=global_args_file, **kwargs)

    xml_content = preprocess_xml(parent_folder, xml_content, resources, load_and_process_file_contents)

    # Post-process the XML
    global_css = css_file.read_text() if css_file is not None else ''
    return _post_process_content(xml_content, global_css)


def get_course(api_token: str, api_url: str, canvas_course_id: int) -> Course:
    """
    Returns a Canvas Course object for the given API URL, API token, and course ID.

    :param api_url: str: The URL for the Canvas API.
    :param api_token: str: The authentication token for the Canvas API.
    :param canvas_course_id: int: The ID of the Canvas course.
    :return: Course: A Canvas Course object.
    """
    canvas = Canvas(api_url, api_token)
    course: Course = canvas.get_course(canvas_course_id)
    return course


def main(
        canvas_api_token: str,
        course_info: CourseInfo,
        input_file: Path,
        args_file: Path = None,
        global_args_file: Path = None,
        line_id: str = None,
        css_file: Path = None,
):
    # Make sure the course actually exists before doing any real effort
    course = get_course(canvas_api_token, course_info['CANVAS_API_URL'], course_info['CANVAS_COURSE_ID'])
    logger = get_logger(course.name)
    logger.info('Connecting to Canvas')

    resources = ResourceManager()

    # Load file
    logger.info('Reading file: ' + str(input_file))
    content_type, content = read_content(input_file)
    processed_content = process_file(
        resources,
        input_file.parent,
        content,
        content_type,
        global_args_file,
        args_file,
        line_id,
        css_file
    )

    # Parse file into XML
    resources = process_canvas_xml(resources, processed_content)

    # Deploy XML
    logger.info('Deploying to Canvas')
    deploy_to_canvas(course, course_info['LOCAL_TIME_ZONE'], resources)


def entry():
    parser = argparse.ArgumentParser()
    # Time zone identifiers: https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    # Use the time zone of the canvas course
    parser.add_argument("--course-info", type=Path, default="canvas_course_info.json")
    parser.add_argument("filename", type=Path)
    parser.add_argument("--args", type=Path, default=None)
    parser.add_argument("--global-args", type=Path, default=None)
    parser.add_argument("--id", type=str, default=None)
    parser.add_argument("--css", type=Path, default=None)
    parser.add_argument('--debug', action='store_true')
    args = parser.parse_args()

    with open(args.course_info) as f:
        course_settings = json.load(f)

    api_token = os.environ.get("CANVAS_API_TOKEN")
    if api_token is None:
        raise ValueError("Please set the CANVAS_API_TOKEN environment variable")

    if args.debug:
        logger.setLevel(logging.DEBUG)

    main(
        canvas_api_token=api_token,
        course_info=course_settings,
        input_file=args.filename,
        args_file=args.args,
        global_args_file=args.global_args,
        line_id=args.id,
        css_file=args.css
    )


if __name__ == '__main__':
    entry()
